<?php

namespace Drupal\dkan_chart\Form;

use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Ajax\SettingsCommand;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\dkan_chart\Command\DkanVisualizeCommand;
use Drupal\dkan_chart\Service\DatastoreVisualizationModeller;
use Drupal\node\Entity\Node;
use Drupal\node\NodeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class VisualizeAjaxUiForm extends \Drupal\Core\Form\FormBase {

  /**
   * @var DatastoreVisualizationModeller
   */
  protected $visualizationModeller;

  /**
   * @var EntityRepositoryInterface
   */
  protected $entityRepository;

  public function __construct(DatastoreVisualizationModeller $visualization_modeller, EntityRepositoryInterface $entity_repository) {
    $this->visualizationModeller = $visualization_modeller;
    $this->entityRepository = $entity_repository;
  }

  public static function create(ContainerInterface $container)
  {
    return new static(
      $container->get('datastore_visualization_modeller'),
      $container->get('entity.repository')
    );
  }

  public function getFormId()
  {
    return 'dkan_visualize_ui';
  }

  public function buildForm(array $form, \Drupal\Core\Form\FormStateInterface $form_state, NodeInterface $node = NULL, $distribution = FALSE)
  {
    $form['mychart'] = [
      '#markup' => '<canvas id="dkanChart"></canvas>',
      '#allowed_tags' => [
        'canvas'
      ]
    ];

    $form['messages'] = [
      '#markup' => '<div class="message-box"></div>'
    ];
    $distribution_node = $distribution ? $this->entityRepository->loadEntityByUuid('node', $distribution) : NULL;
    $this->visualizationModeller->initData($node, FALSE, $distribution_node);
    if ($this->visualizationModeller->getDistribution()) {
      $form['distribution_label'] = [
        '#type' => 'markup',
        '#markup' => $this->visualizationModeller->getDistributionJSON()->data->title
      ];
    }
    $form['chart_type'] = [
      '#type' => 'select',
      '#title' => t('Type'),
      '#options' => [
        'barx' => t('Vertical bars'),
        'bary' => t('Horizontal bars'),
        'barxstack' => t('Vertical bars stacked'),
        'barystack' => t('Horizontal bars stacked'),
        'pie' => t('Pie'),
        'doughnut' => t('Doughnut'),
        'line' => t('Line')
      ],
      '#required' => TRUE,
    ];
    $form['calculation'] = [
      '#type' => 'select',
      '#title' => t('Calculation'),
      '#options' => [
        'sum' => t('Sum'),
        'count' => t('Count'),
        'min' => t('Minimum'),
        'max' => t('Maximum'),
        'avg' => t('Average'),
      ],
      '#required' => TRUE,
      '#default_value' => 'sum'
    ];
    $form['label'] = [
      '#type' => 'select',
      '#title' => t('Label'),
      '#options' => $this->visualizationModeller->getFieldsOptions(),
      '#required' => TRUE,
    ];
    $form['data'] = [
      '#type' => 'select',
      '#title' => t('Data'),
      '#multiple' => TRUE,
      '#options' => $this->visualizationModeller->getFieldsOptions(['numeric-date', 'numeric', 'leading-numeric']),
      '#required' => TRUE,
      '#attributes' => [
        'class' => [
          'choice-select'
        ],
        'data-placeholdervalue' => $this->t('Choose data')
      ]
    ];
    $form['series'] = [
      '#type' => 'select',
      '#title' => t('Series'),
      '#options' => $this->visualizationModeller->getFieldsOptions(),
      '#empty_value' => FALSE
    ];
    $form['filter'] = [
      '#type' => 'select',
      '#multiple' => TRUE,
      '#title' => t('Filter'),
      '#options' => $this->visualizationModeller->filterOptions(),
      '#attributes' => [
        'class' => [
          'choice-select'
        ],
        'data-placeholdervalue' => $this->t('Choose filters'),
        'data-addgroupinsearch' => 'true',
      ]
    ];
    $form['node'] = [
      '#type' => 'hidden',
      '#value' => $node->id()
    ];

    $form['actions'] = [
      '#type' => 'button',
      '#value' => $this->t('Visualize'),
      '#ajax' => [
        'callback' => '::visualize'
      ]
    ];

    return $form;
  }

  public function visualize(array $form, \Drupal\Core\Form\FormStateInterface $formState) {
    $response = new \Drupal\Core\Ajax\AjaxResponse();

    if ($formState->getErrors()) {
      $error_content = [
        '#type' => 'status_messages'
      ];
      $response->addCommand(new HtmlCommand('.message-box', $error_content));
      return $response;
    }

    $chart_type = $formState->getValue('chart_type');
    $this->visualizationModeller->setChartType($chart_type);
    $calculation = $formState->getValue('calculation');
    $label = $formState->getValue('label');
    $data = $formState->getValue('data');
    $series = $formState->getValue('series');
    $node_id = $formState->getValue('node');
    $filter =  $formState->getValue('filter');
    $filter_options = [];
    foreach ($filter as $filter_key => $filter_description) {
      $filter_parts = explode(':', $filter_key);
      $filter_options[$filter_parts[0]][] = $filter_parts[1];
    }

    $node = Node::load($node_id);
    // Clear message box from subsequent errors
    $response->addCommand(new HtmlCommand('.message-box', []));

    $response->addCommand(new SettingsCommand(['chart_dataset' => null], TRUE));
    $response->addCommand(new SettingsCommand(['chart_dataset' => $this->visualizationModeller->uniqueValuesMultiple($label, array_values($data), $series, $calculation, $filter_options)], TRUE));
    $response->addCommand(new SettingsCommand(['chart_definition' => null], TRUE));
    $response->addCommand(new SettingsCommand(['chart_definition' => $this->visualizationModeller->getChartDefinition()], TRUE));
    $response->addCommand(
      new DkanVisualizeCommand('dkanVisualize', [
        'label' => $label,
        'datacol' => $data
      ])
    );

    return $response;
  }

  public function submitForm(array &$form, \Drupal\Core\Form\FormStateInterface $form_state)
  {
    // not called, it's Ajax
  }
}
