<?php

namespace Drupal\dkan_chart\Service;

use Drupal\metastore\MetastoreService;
use Drupal\node\NodeInterface;

class DatastoreVisualizationModeller {

  protected $metastore;

  protected $data;

  /**
   * @var NodeInterface
   */
  protected $node;

  /**
   * @var NodeInterface|null
   */
  protected $distribution;

  protected $resource_id = 0;

  protected $chartType;

  protected $fields;

  protected $chartDefinition = [
    'barx' => [
      'chart' => 'bar',
      'axis' => 'x',
      'axiskey' => 'y',
      'stacked' => FALSE,
    ],
    'bary' => [
      'chart' => 'bar',
      'axis' => 'y',
      'axiskey' => 'x',
      'stacked' => FALSE,
    ],
    'barxstack' => [
      'chart' => 'bar',
      'axis' => 'x',
      'axiskey' => 'y',
      'stacked' => TRUE,
    ],
    'barystack' => [
      'chart' => 'bar',
      'axis' => 'y',
      'axiskey' => 'x',
      'stacked' => TRUE,
    ],
    'pie' => [
      'chart' => 'pie',
    ],
    'doughnut' => [
      'chart' => 'doughnut',
    ],
    'line' => [
      'chart' => 'line',
      'axis' => 'x',
      'axiskey' => 'y',
    ],
  ];

  public function __construct(MetastoreService $metastore) {
    $this->metastore = $metastore;
  }

  /**
   * @return array
   * @throws \Drupal\metastore\Exception\InvalidJsonException
   */
  public function getDistributionsInDatastore() {
    $data = $this->metastore->swapReferences($this->metastore->get('dataset', $this->node->uuid()));
    $distributions = $data->{"$.distribution"};
    $identifiers = [];
    foreach ($distributions as $dist) {
      try {
        $response =\Drupal::httpClient()->request(
          'GET',
          \Drupal::request()->getSchemeAndHttpHost() . '/api/1/datastore/query/' . $dist['identifier'] . '?limit=1'
        );
        if ($response->getStatusCode() == '200') {
          // ToDo: Refactor to Switch key and value, beware of uses, one is with array_flip
          $identifiers[$dist['data']['title']] = $dist['identifier'];
        }
      }
      catch (\Exception $exception) {
        // Skip this distribution
      }
    }
    return $identifiers;
  }

  public function setNode(NodeInterface $node) {
    $this->node = $node;
  }

  public function setDistribution(?NodeInterface $distribution) {
    $this->distribution = $distribution;
  }

  public function getDistribution() {
    return $this->distribution;
  }

  public function getDistributionJSON() {
    if ($this->distribution && $this->distribution->hasField('field_json_metadata') && !$this->distribution->get('field_json_metadata')->isEmpty()) {
      return json_decode($this->distribution->get('field_json_metadata')->value);
    }
    return FALSE;
  }

  public function initData(NodeInterface $node, bool $limit, NodeInterface $distribution = NULL)
  {
    $this->setNode($node);
    $this->setDistribution($distribution);
    $this->setData($limit);
    $this->setFields();
  }

  protected function setFields()
  {
    $fields = $this->data->schema->{$this->getDistributionId()}->fields ?? [];
    foreach ($fields as $api_name => $field) {
      $this->fields[$api_name] = [
        'description' => $field->description ?? $api_name,
        'type' => $this->estimateFieldType($api_name),
      ];
    }
  }

  public function getFields() {
    return $this->fields;
  }

  public function setChartType(string $chart_type)
  {
    $this->chartType = $chart_type;
  }

  protected function getChartType() {
    return $this->chartType ?? 'xbar';
  }

  public function getFieldsOptions(array $type = []) {
    $fields_options = [];
    foreach ($this->getFields() as $api_field_name => $field) {
      if (empty($type) || in_array($field['type'], $type)) {
        $fields_options[$api_field_name] = $field['description'];
      }
    }
    return $fields_options;
  }

  public function filterOptions()
  {
    $filter_options = [];
    $field_options = $this->getFieldsOptions();
    foreach ($field_options as $column_name => $column_description) {
      $filter_options[$column_description] = [];
      foreach ($this->data->results as $result) {
        $filter_options[$column_description][$column_name . ':' . $result->{$column_name}] = $result->{$column_name};
      }
    }
    return $filter_options;
  }

  public function uniqueValuesMultiple(string $column_name, array $sum_column, string $dim_column, string $calc_type, array $filters = [], int $limit_items = 100) {
    $datasets = [];
    foreach ($sum_column as $sum) {
      $datasets[$sum] = $this->uniqueValues($column_name, $sum, $dim_column, $filters, $limit_items, $calc_type);
    }
    $chart_def_label = $this->getChartDefinition()['label'];
    $chart_label = [];
    foreach ($sum_column as $sum) {
      $chart_label[$sum] = $chart_def_label;

    }
    $this->setChartDefintionLabels($chart_label);
    return $datasets;
  }

  protected function valueCalculation(&$unique_values, $result, $column_name, $sum_column, $result_dim, $calc_type) {
    if (empty($unique_values[$result->{$column_name}][$result_dim])) {
      switch ($calc_type) {
        case 'count':
          $unique_values[$result->{$column_name}][$result_dim] = 1;
          break;
        case 'sum':
        case 'min':
        case 'max':
          $unique_values[$result->{$column_name}][$result_dim] = floatval($result->{$sum_column});
          break;
        case 'avg':
          $unique_values[$result->{$column_name}][$result_dim] = [
            'count' => 1,
            'sum' => floatval($result->{$sum_column})
          ];
      }
    } else {
      switch ($calc_type) {
        case 'count':
          $unique_values[$result->{$column_name}][$result_dim]++;
          break;
        case 'sum':
          $unique_values[$result->{$column_name}][$result_dim] += floatval($result->{$sum_column});
          break;
        case 'min':
          if ($unique_values[$result->{$column_name}][$result_dim] > floatval($result->{$sum_column})) {
            $unique_values[$result->{$column_name}][$result_dim] = floatval($result->{$sum_column});
          }
          break;
        case 'max':
          if ($unique_values[$result->{$column_name}][$result_dim] < floatval($result->{$sum_column})) {
            $unique_values[$result->{$column_name}][$result_dim] = floatval($result->{$sum_column});
          }
          break;
        case 'avg':
          $unique_values[$result->{$column_name}][$result_dim]['count']++;
          $unique_values[$result->{$column_name}][$result_dim]['sum'] += floatval($result->{$sum_column});
          break;
      }
    }
  }

  protected function uniqueValues(string $column_name, string $sum_column, string $dim_column, array $filters, int $limit_items, string $calc_type)
  {
    if ($dim_column) {
      if ($limit_items) {
        $limit_items = ceil($limit_items / count($this->unigueValues($dim_column)));
        $this->setChartDefintionLabels(array_slice($this->unigueValues($dim_column), 0, $limit_items));
      } else {
        $this->setChartDefintionLabels($this->unigueValues($dim_column));
      }
    }
    else {
        $this->setChartDefintionLabels(['default']);
    }

    $unique_values = [];

    foreach ($this->data->results as $result) {
      if ($this->filterValues($result, $filters)) {
        $result_dim = $result->{$dim_column} ?? 'default';
        $this->valueCalculation($unique_values, $result, $column_name, $sum_column, $result_dim, $calc_type);
      }
    }
    if ($calc_type === 'avg') {
      foreach ($unique_values as $unique_key => $unique_value) {
        foreach ($unique_value as $uniq_key => $uniq_val) {
          $unique_values[$unique_key][$uniq_key] = $uniq_val['sum'] / $uniq_val['count'];
        }
      }
    }

    arsort($unique_values, SORT_NUMERIC);
    if ($limit_items) {
      $unique_values = array_slice($unique_values, 0, $limit_items, TRUE);
    }

    $datasets = [];

    if ($this->getChartType() === 'pie' || $this->getChartType() === 'doughnut') {
      $this->setChartDefintionLabels([]);
      foreach ($unique_values as $dim => $dataset) {
        $this->addChartDefintionLabel($dim);
        foreach ($dataset as $x => $y) {
          $datasets[$x][$dim] = $y;
        }
      }
      foreach ($unique_values as $unique_key => $unique_value) {
        foreach ($datasets as $dataset_key => $dataset) {
          if (!isset($dataset[$unique_key])) {
            $datasets[$dataset_key][$unique_key] = 0;
          }
        }
      }
      foreach ($datasets as $dataset_key => $dataset) {
        ksort($dataset);
        $datasets[$dataset_key] = $dataset;
      }
      $datasets_plain = [];
      foreach ($datasets as $dim => $dataset) {
        $datasets_plain[$dim] = array_values($dataset);
      }

      $labels = $this->getChartDefinition()['label'];
      sort($labels);
      $this->setChartDefintionLabels($labels);

      return $datasets_plain;
    }
    foreach ($unique_values as $dim => $dataset) {
      $data_object = new \stdClass();
      $data_object->{$this->chartDefinition[$this->getChartType()]['axis']} = (string) $dim;
      foreach ($dataset as $x => $y) {
        $data_object->{(string) $x} = $y;
      }
      $datasets[] = $data_object;
    }
    return $datasets;
  }

  protected function estimateFieldType ($field_name) {
    $tests = array_rand($this->data->results, count($this->data->results) < 5 ? count($this->data->results): 5);
    $tested = [];
    foreach ($tests as $test) {
      $tested[] = [
        'numeric' => is_numeric($this->data->results[$test]->{$field_name}),
        'date' => strtotime($this->data->results[$test]->{$field_name}),
        'leading_numeric' => preg_match('/(?m)^(\d+).*/m', $this->data->results[$test]->{$field_name}) && !is_numeric($this->data->results[$test]->{$field_name}),
      ];
    }
    $numeric = TRUE;
    foreach ($tested as $round) {
      $numeric = $round['numeric'];
      if ($numeric === FALSE) {
        break;
      }
    }
    $leading_numeric = TRUE;
    foreach ($tested as $round) {
      $leading_numeric = $round['leading_numeric'];
      if ($leading_numeric === FALSE) {
        break;
      }
    }
    $date = TRUE;
    $epoch = [];
    foreach ($tested as $round) {
      $date = $round['date'];
      if ($date === FALSE) {
        break;
      }
      $epoch[] = $date === strtotime(0);
    }
    $all_epoch = TRUE;
    foreach ($epoch as $is_epoch) {
      if (!$is_epoch) {
        $all_epoch = FALSE;
        break;
      }
    }
    $date = $date && $all_epoch;

    if ($numeric && $date) {
      return 'numeric-date';
    }
    if ($numeric) {
      return 'numeric';
    }
    if ($date) {
      return 'date';
    }
    if ($leading_numeric) {
      return 'leading-numeric';
    }
    return 'text';
  }

  public function getChartDefinition() {
    $this->chartDefinition['type'] = $this->getChartType();
    return $this->chartDefinition;
  }

  protected function filterValues(object $values, array $filters)
  {
    foreach ($filters as $filter_col => $filter_values) {
      if (!in_array($values->{$filter_col}, $filter_values)) {
        return FALSE;
      }
    }
    return TRUE;
  }

  protected function getDatastoreApiUrl($limit = FALSE)
  {
    if ($this->node instanceof NodeInterface) {
      if (!empty($distributions = $this->getDistributionsInDatastore())) {
        $distribution = $this->distribution ? $this->distribution->uuid() : reset($distributions);
        $base_datastore_api_url = \Drupal::request()
            ->getSchemeAndHttpHost() . '/api/1/datastore/query/' . $distribution;
        if ($limit === FALSE) {
          $base_datastore_api_url = $base_datastore_api_url . '?limit=0';
        }
        return $base_datastore_api_url;
      }
    }
    return FALSE;
  }

  protected function getDistributionId()
  {
    return $this->data->query->resources[$this->resource_id]->id;
  }

  protected function setChartDefintionLabels(array $labels)
  {
    $this->chartDefinition['label'] = $labels;
  }

  protected function addChartDefintionLabel(string $label) {
    $this->chartDefinition['label'][] = $label;
  }


  protected function setData($limit = FALSE)
  {
    if ($this->getDatastoreApiUrl($limit)) {
      $this->data = json_decode(file_get_contents($this->getDatastoreApiUrl($limit)));
    }
  }

  protected function unigueValues(string $column_name)
  {
    $unique_values = [];
    foreach ($this->data->results as $result) {
      $unique_values[$result->{$column_name}] = TRUE;
    }
    return array_keys($unique_values);
  }

}
